/*
 * @(#)CalcParser.java
 */
// import java.io.*;
import java.io.StringReader;
import java.io.PushbackReader;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.IOException;

// import java.util.*;
import java.util.List;
import java.util.ArrayList;

/**
 * @see CreamSolver
 * @since 1.0
 * @version 1.0, 05/01/14
 * @author Naoyuki Tamura
 */
public class CalcParser {
    private PushbackReader reader_;
    private Token token_ = null;

    /**
     * 
     */
    public class Token {
	public String type;
	public String str;

	public String toString() {
	    return str;
	}
    }

    /**
     *
     */
    public class Expression {
	public Token op = null;
	public Expression[] args = null;

	public Expression(Token op) {
	    this.op = op;
	    args = null;
	}

	public Expression(Token op, Expression arg1) {
	    this.op = op;
	    args = new Expression[1];
	    args[0] = arg1;
	}

	public Expression(Token op, Expression arg1, Expression arg2) {
	    this.op = op;
	    args = new Expression[2];
	    args[0] = arg1;
	    args[1] = arg2;
	}

	public Expression(Token op, List<Expression> args) {
	    this.op = op;
	    this.args = null;
	    if (args != null && args.size() > 0) {
		this.args = new Expression[1];
		this.args = args.toArray(this.args);
	    }
	}
	
	public Expression(Token op, Expression[] args) {
	    this.op = op;
	    this.args = args;
	}

	public int argc() {
	    return args == null ? 0 : args.length;
	}

	public boolean isInteger() {
	    return op.type.equals("INTEGER");
	}

	public boolean isFunctor(String name) {
	    return op.type.equals("SYMBOL") && op.str.equals(name);
	}

	public boolean isFunctor(String name, int arity) {
	    return isFunctor(name) && argc() == arity;
	}

	public boolean isSymbol(String name) {
	    return isFunctor(name, 0);
	}

	public boolean isSymbol() {
	    return op.type.equals("SYMBOL") && argc() == 0;
	}

	public boolean isOperator(String name) {
	    return op.type.equals(name);
	}

	public boolean isOperator(String name, int arity) {
	    return isOperator(name) && argc() == arity;
	}

	public String toString() {
	    String s = "";
	    if (args == null) {
		s = op.toString();
	    } else {
		String delim = "";
		for (int i = 0; i < args.length; i++) {
		    s += delim + args[i].toString();
		    delim = ";";
		}
		s = op.toString() + "(" + s + ")";
	    }
	    return s;
	}
    }

    /**
     * Constructor.
     * @param inputString 
     */
    public CalcParser(String inputString) {
	this(new PushbackReader(new StringReader(inputString)));
    }

    /**
     * Constructor.
     * @param stream
     */
    public CalcParser(InputStream stream) throws IOException {
	this(new BufferedReader(new InputStreamReader(stream)).readLine());
    }

    /**
     * Constructor.
     * @param reader 
     */
    public CalcParser(PushbackReader reader) {
	this.reader_ = reader;
	token_ = null;
    }

    private boolean _isWhitespace(int c) {
	return Character.isWhitespace((char)c);
    }

    private boolean _isDigit(int c) {
	return Character.isDigit((char)c);
    }

    private boolean _isSymbolStart(int c) {
	return (Character.isUnicodeIdentifierStart((char)c) || c == '$')
	    && ! Character.isIdentifierIgnorable((char)c);
    }

    private boolean _isSymbolPart(int c) {
	return (Character.isUnicodeIdentifierPart((char)c) || 
		c == '$' || c == ':' || c == '.')
	    && ! Character.isIdentifierIgnorable((char)c);
    }

    private void _unread(int c) throws IOException {
	if (c >= 0) {
	    reader_.unread(c);
	}
    }

    /**
     * 
     * @return nextToken.
     */
    public Token nextToken() {
	Token token = null;
	StringBuffer type = new StringBuffer();
	StringBuffer str = new StringBuffer();
	try {
	    int c = reader_.read();
	    if (c < 0) {
		return null;
	    } else if (_isWhitespace(c)) {
		type = new StringBuffer(" ");
		while  (_isWhitespace(c)) {
		    c = reader_.read();
		}
		_unread(c);
	    } else if (_isDigit(c)) {
		type = new StringBuffer("INTEGER");
		do {
		    str.append((char)c);
		    c = reader_.read();
		} while (_isDigit(c));
		_unread(c);
	    } else if (c == '"') {
		type = new StringBuffer("STRING");
		while (true) {
		    c = reader_.read();
		    if (c == '"') {
			c = reader_.read();
			if (c != '"') {
			    _unread(c);
			    break;
			}
		    }
		    str.append((char)c);
		}
	    } else if (_isSymbolStart(c)) {
		type = new StringBuffer("SYMBOL");
		do {
		    str.append(Character.toUpperCase((char)c));
		    c = reader_.read();
		} while (_isSymbolPart(c));
		_unread(c);
	    } else if (c == '<') {
		type.append((char)c);
		str.append((char)c);
		c = reader_.read();
		if (c == '=' || c == '>') {
		    type.append((char)c);
		    str.append((char)c);
		} else {
		    _unread(c);
		}
	    } else if (c == '>') {
		type.append((char)c);
		str.append((char)c);
		c = reader_.read();
		if (c == '=') {
		    type.append((char)c);
		    str.append((char)c);
		} else {
		    _unread(c);
		}
	    } else {
		type.append((char)c);
		str.append((char)c);
	    }
	    token = new Token();
	    token.type = type.toString();
	    token.str = str.toString();
	} catch (IOException e) { e.printStackTrace(); }
	return token;
    }

    private void _skipSpaces() {
	while (token_ != null && token_.type.equals(" ")) {
	    token_ = nextToken();
	}
    }

    /**
     * Expression ::= Expr { CompOp Expr }
     * CompOp ::= '=' | '<>' | '<' | '<=' | '>' | '>='
     */
    private Expression _parseExpression() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	x = _parseExpr();
	for (_skipSpaces(); token_ != null; _skipSpaces()) {
	    if (token_.type.equals("=") || token_.type.equals("<>")
		|| token_.type.equals("<") || token_.type.equals("<=")
		|| token_.type.equals(">") || token_.type.equals(">=")) {
		op = token_;
		token_ = nextToken();
		y = _parseExpr();
		x = new Expression(op, x, y);
	    } else {
		break;
	    }
	}
	return x;
    }

    /**
     * Expr ::= [ '+' | '-' ] Term { ( '+' | '-' ) Term }
     */
    private Expression _parseExpr() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	_skipSpaces();
	if (token_ == null) {
	    return null;
	} else if (token_.type.equals("+") || token_.type.equals("-")) {
	    op = token_;
	    token_ = nextToken();
	    y = _parseTerm();
	    x = new Expression(op, y);
	} else {
	    x = _parseTerm();
	}
	for (_skipSpaces(); token_ != null; _skipSpaces()) {
	    if (token_.type.equals("+") || token_.type.equals("-")) {
		op = token_;
		token_ = nextToken();
		y = _parseTerm();
		x = new Expression(op, x, y);
	    } else {
		break;
	    }
	}
	return x;
    }

    /**
     * Term ::= Factor { ( '*' | '/' ) Factor }
     */
    private Expression _parseTerm() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	x = _parseFactor();
	for (_skipSpaces(); token_ != null; _skipSpaces()) {
	    if (token_.type.equals("*") || token_.type.equals("/")) {
		op = token_;
		token_ = nextToken();
		y = _parseFactor();
		x = new Expression(op, x, y);
	    } else {
		break;
	    }
	}
	return x;
    }

    /**
     * Factor ::= Integer | String | Symbol |
     *            Symbol '(' Expression { ';' Expression } ')' |
     *            '(' Expression ')'
     */
    private Expression _parseFactor() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	_skipSpaces();
	if (token_ == null) {
	    return null;
	} else if (token_.type.equals("INTEGER") || 
		   token_.type.equals("STRING")) {
	    op = token_;
	    token_ = nextToken();
	    x = new Expression(op);
	} else if (token_.type.equals("SYMBOL")) {
	    op = token_;
	    token_ = nextToken();
	    _skipSpaces();
	    if (token_ != null && token_.type.equals("(")) {
		ArrayList<Expression> args = new ArrayList<Expression>();
		do {
		    token_ = nextToken();
		    y = _parseExpression();
		    args.add(y);
		    _skipSpaces();
		} while (token_ != null && token_.type.equals(";"));
		if (token_ == null || ! token_.type.equals(")")) {
		    throw new IllegalArgumentException
			(") is expected at " + token_);
		}
		token_ = nextToken();
		x = new Expression(op, args);
/*
	    } else if (token_ != null && token_.type.equals(":")) {
		x = new Expression(op);
		op = token_;
		token_ = nextToken();
		y =  _parseFactor();
		x = new Expression(op, x, y);
*/
	    } else {
		x = new Expression(op);
	    }
	} else if (token_.type.equals("(")) {
	    token_ = nextToken();
	    x = _parseExpression();
	    _skipSpaces();
	    if (token_ == null || ! token_.type.equals(")")) {
		throw new IllegalArgumentException
		    (") is expected at " + token_);
	    }
	    token_ = nextToken();
	} else {
	    throw new IllegalArgumentException
		("can not start expression at " + token_);
	}
	return x;
    }

    /**
     * Parse a calc expression. 
     * @return 
     */
    public Expression parse() throws IllegalArgumentException {
	token_ = nextToken();
	_skipSpaces();
	if (token_.type.equals("=")) {
	    token_ = nextToken();
	}
	Expression x = _parseExpression();
	_skipSpaces();
	if (token_ != null) {
	    throw new IllegalArgumentException
		("unknown operator at " + token_);
	}
	return x;
    }
    
    public static void main(String[] args) {
	CalcParser parser;
	try {
	    while (true) {
		parser = new CalcParser(System.in);
		
		Expression expr = parser.parse();
		System.out.println(expr);
	    }
	} catch (IOException e) { e.printStackTrace();
	} catch (IllegalArgumentException e) { e.printStackTrace(); }
    }
}
